//********************************************************************************************************************************
//*                                                                                                                              *
//*     TMultiThread 2.0.7.8  3delite 2011-2024                                                                                 *
//*     See TMultiThread2 ReadMe.txt for details.                                                                                *
//*                                                                                                                              *
//* Licenses available for usage of this component:                                                                              *
//* Unlimited Freeware License: 15                                                                                              *
//*     http://www.shareit.com/product.html?productid=300960106                                                                  *
//* Unlimited Shareware License: 25                                                                                             *
//*     http://www.shareit.com/product.html?productid=300432175                                                                  *
//* Single Commercial License: 45                                                                                               *
//*     http://www.shareit.com/product.html?productid=300432176                                                                  *
//*                                                                                                                              *
//*     https://www.3delite.hu/Object Pascal Developer Resources/TMultiThread.html                                               *
//*                                                                                                                              *
//* For other Delphi components see the home page:                                                                               *
//*                                                                                                                              *
//*     https://www.3delite.hu/                                                                                                  *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit MultiThread2;

interface

{$WARN SYMBOL_PLATFORM OFF}

Uses
    Classes,
    SysUtils,
    System.SyncObjs,
    System.Generics.Collections;

type
    TMultiThreadMessageQueue = class
    private
        FTObjectQueue: TObjectQueue<TObject>;
        FCriticalSection: TCriticalSection;
    public
        Constructor Create;
        Destructor Destroy; override;
        procedure Add(Obj: TObject);
        function Pop: TObject;
        function Clear: Integer;
        function Count: Integer;
        procedure SetCapacity(Value: Integer);
    end;

type
    TMultiThread2 = class;

    TMultiThreadThread = class;

    TThreadStartCallback = procedure(Thread: TMultiThreadThread; var Parameters: TObject) of object;
    TThreadFinishedCallback = procedure(Thread: TMultiThreadThread; var Parameters: TObject) of object;
    TThreadWorkerCallback = procedure(Thread: TMultiThreadThread; var Parameters: TObject; var Job: TObject; var Results: TObject) of object;
    TThreadResultsCallback = procedure(Sender: TObject; var Results: TObject) of object;
    TThreadMessage = procedure(Sender: TObject; var MessageObject: TObject) of object;

    TThreadCount = (tcAutomaticMax, tcAutomaticMaxMinus1, tcAutomaticAllCores, tcAutomaticAllCoresMinus1, tcManual);

    TMultiThreadSettings = record
        WorkerCallback: TThreadWorkerCallback;
        {$IFDEF MSWINDOWS}
        ThreadPriority: TThreadPriority;
        {$ENDIF}
        ThreadStartCallback: TThreadStartCallback;
        ThreadFinishedCallback: TThreadFinishedCallback;
        WorkMessageQueue: TMultiThreadMessageQueue;
        MessageQueue: TMultiThreadMessageQueue;
        ResultMessageQueue: TMultiThreadMessageQueue;
        Parameters: TObject;
        AutoFreeData: Boolean;
        AutoFreeResults: Boolean;
        Parent: TMultiThread2;
    end;

    TMultiThreadThread = class(TThread)
    private
        FSettings: TMultiThreadSettings;
        FResults: TObject;
        procedure SendResults;
    public
        Data: TObject;
        Constructor Create(ThreadSettings: TMultiThreadSettings);
        Destructor Destroy; override;
        procedure SendMessage(MessageObject: TObject);
        property Terminated;
    protected
        procedure Execute; override;
    end;

    [ComponentPlatformsAttribute(pidWin32 OR pidWin64 OR pidOSX32 {$IF CompilerVersion >= 33} OR pidiOSSimulator32 OR pidiOSDevice32 OR {$IF CompilerVersion > 34} pidAndroidArm32 {$ELSE} pidAndroid32Arm {$ENDIF} OR pidLinux64 {$ELSE} {$IF CompilerVersion >= 26} OR pidiOSSimulator OR pidiOSDevice OR pidAndroid {$ENDIF} {$ENDIF})]
    TMultiThread2 = class(TComponent)
    private
        FThreadCount: TThreadCount;
        FThreadCountManual: Integer;
        FThreadCountUse: Integer;
        FAutoFreeData: Boolean;
        FAutoFreeResults: Boolean;
        FAutoFreeMessages: Boolean;
        FAutoFreeParameters: Boolean;
        EThreadStartCallback: TThreadStartCallback;
        EThreadFinishedCallback: TThreadFinishedCallback;
        EThreadWorkerCallback: TThreadWorkerCallback;
        EThreadResultsCallback: TThreadResultsCallback;
        EThreadMessage: TThreadMessage;
        FThreads: Array of TMultiThreadThread;
        FWorkToDo: UInt64;
        function GetWorkLeft: NativeInt;
        function GetWorkFinished: Boolean;
    public
        {$IFDEF MSWINDOWS}
        ThreadPriority: TThreadPriority;
        {$ENDIF}
        Parameters: TObject;
        WorkMessageQueue: TMultiThreadMessageQueue;
        MessageQueue: TMultiThreadMessageQueue;
        ResultMessageQueue: TMultiThreadMessageQueue;
        Constructor Create(AOwner: TComponent); override;
        Destructor Destroy; override;
        procedure Start;
        procedure Stop;
        //procedure Pause;
        function AddWork(Job: TObject): Boolean;
        function Active: Boolean;
        procedure Clear;
        procedure TerminateAllThreads;
        procedure CheckIncomingMessages;
        procedure CheckIncomingResults;
        property WorkLeft: NativeInt read GetWorkLeft;
        property WorkFinished: Boolean read GetWorkFinished;
        property ThreadsRunningCount: Integer read FThreadCountUse;
    published
        property ThreadCount: TThreadCount read FThreadCount write FThreadCount default tcAutomaticMax;
        property ThreadCountManual: Integer read FThreadCountManual write FThreadCountManual default 4;
        property AutoFreeData: Boolean read FAutoFreeData write FAutoFreeData default True;
        property AutoFreeResults: Boolean read FAutoFreeResults write FAutoFreeResults default True;
        property AutoFreeMessages: Boolean read FAutoFreeMessages write FAutoFreeMessages default True;
        property AutoFreeParameters: Boolean read FAutoFreeParameters write FAutoFreeParameters default True;
        property ThreadStartCallback: TThreadStartCallback read EThreadStartCallback write EThreadStartCallback;
        property ThreadFinishedCallback: TThreadFinishedCallback read EThreadFinishedCallback write EThreadFinishedCallback;
        property ThreadWorkerCallback: TThreadWorkerCallback read EThreadWorkerCallback write EThreadWorkerCallback;
        property ThreadResultsCallback: TThreadResultsCallback read EThreadResultsCallback write EThreadResultsCallback;
        property ThreadMessage: TThreadMessage read EThreadMessage write EThreadMessage;
    end;

    procedure Register;

implementation

Uses
    LogicalProcessorInformation;

procedure Register;
begin
    RegisterComponents('3delite', [TMultiThread2]);
end;


{ TThreadMessageQue }

procedure TMultiThreadMessageQueue.Add(Obj: TObject);
begin
    FCriticalSection.Enter;
    try
        FTObjectQueue.Enqueue(Obj);
    finally
        FCriticalSection.Leave;
    end;
end;

function TMultiThreadMessageQueue.Clear: Integer;
var
    i: Integer;
    Obj: TObject;
begin
    FCriticalSection.Enter;
    try
        Result := 0;
        for i := FTObjectQueue.Count - 1 downto 0 do begin
            Obj := FTObjectQueue.Extract;
            if Assigned(Obj) then begin
                {$IFDEF NEXTGEN}
                Obj.DisposeOf;
                {$ELSE}
                FreeAndNil(Obj);
                {$ENDIF}
                Inc(Result);
            end;
        end;
    finally
        FCriticalSection.Leave;
    end;
end;

constructor TMultiThreadMessageQueue.Create;
begin
    FTObjectQueue := TObjectQueue<TObject>.Create;
    FTObjectQueue.OwnsObjects := False;
    FTObjectQueue.Capacity := 256;
    FCriticalSection := TCriticalSection.Create;
end;

destructor TMultiThreadMessageQueue.Destroy;
begin
    FreeAndNil(FCriticalSection);
    FreeAndNil(FTObjectQueue);
    inherited;
end;

function TMultiThreadMessageQueue.Pop: TObject;
begin
    FCriticalSection.Enter;
    try
        Result := nil;
        if FTObjectQueue.Count > 0 then begin
            Result := FTObjectQueue.Extract;
        end;
    finally
        FCriticalSection.Leave;
    end;
end;

procedure TMultiThreadMessageQueue.SetCapacity(Value: Integer);
begin
    FTObjectQueue.Capacity := Value;
end;

function TMultiThreadMessageQueue.Count: Integer;
begin
    FCriticalSection.Enter;
    try
        Result := FTObjectQueue.Count;
    finally
        FCriticalSection.Leave;
    end;
end;

Constructor TMultiThreadThread.Create(ThreadSettings: TMultiThreadSettings);
begin
    Inherited Create(False);
    {$IFDEF MSWINDOWS}
    Self.Priority := ThreadSettings.ThreadPriority;
    {$ENDIF}
    FSettings := ThreadSettings;
    FreeOnTerminate := False;
end;

Destructor TMultiThreadThread.Destroy;
begin
    Inherited Destroy;
end;

procedure TMultiThreadThread.Execute;
var
    Work: TObject;
begin
    try
        if Assigned(FSettings.ThreadStartCallback) then begin
            FSettings.ThreadStartCallback(Self, FSettings.Parameters);
        end;
        while NOT Terminated do begin
            try
                Work := FSettings.WorkMessageQueue.Pop;
                if Assigned(Work) then begin
                    if Assigned(FSettings.WorkerCallback) then begin
                        try
                            FResults := nil;
                            FSettings.WorkerCallback(Self, FSettings.Parameters, Work, FResults);
                        except
                            //*
                        end;
                    end;
                    if FSettings.AutoFreeData then begin
                        if Assigned(Work) then begin
                            FreeAndNil(Work);
                        end;
                    end;
                    if (NOT Terminated)
                    AND Assigned(FResults)
                    then begin
                        SendResults;
                    end else begin
                        if FSettings.AutoFreeResults then begin
                            if Assigned(FResults) then begin
                                FreeAndNil(FResults);
                            end;
                        end;
                    end;
                    AtomicDecrement(FSettings.Parent.FWorkToDo);
                end else begin
                    Sleep(1);
                end;
            except
                //*
            end;
        end;
        if Assigned(FSettings.ThreadFinishedCallback) then begin
            FSettings.ThreadFinishedCallback(Self, FSettings.Parameters);
        end;
    except
        //*
    end;
end;

procedure TMultiThreadThread.SendResults;
begin
    FSettings.ResultMessageQueue.Add(FResults);
    FResults := nil;
end;

procedure TMultiThreadThread.SendMessage(MessageObject: TObject);
begin
    if Assigned(MessageObject) then begin
        FSettings.MessageQueue.Add(MessageObject);
    end;
end;

procedure TMultiThread2.Start;
var
    i: Integer;
    ThreadSettings: TMultiThreadSettings;
begin
    Stop;
    //* Determine number of threads to spawn
    case FThreadCount of
        tcAutomaticMax: begin
            FThreadCountUse := ProcessorInformation.LogicalProcessorCount;
        end;
        tcAutomaticMaxMinus1: begin
            FThreadCountUse := ProcessorInformation.LogicalProcessorCount - 1;
        end;
        tcAutomaticAllCores: begin
            FThreadCountUse := ProcessorInformation.ProcessorCoreCount;
        end;
        tcAutomaticAllCoresMinus1: begin
            FThreadCountUse := ProcessorInformation.ProcessorCoreCount - 1;
        end;
        tcManual: begin
            FThreadCountUse := FThreadCountManual;
        end;
    end;
    if FThreadCountUse < 1 then begin
        FThreadCountUse := 1;
    end;
    //* Settings
    ThreadSettings.WorkerCallback := EThreadWorkerCallback;
    {$IFDEF MSWINDOWS}
    ThreadSettings.ThreadPriority := ThreadPriority;
    {$ENDIF}
    ThreadSettings.ThreadStartCallback := EThreadStartCallback;
    ThreadSettings.ThreadFinishedCallback := EThreadFinishedCallback;
    ThreadSettings.WorkMessageQueue := WorkMessageQueue;
    ThreadSettings.MessageQueue := MessageQueue;
    ThreadSettings.ResultMessageQueue := ResultMessageQueue;
    ThreadSettings.Parameters := Parameters;
    ThreadSettings.AutoFreeData := FAutoFreeData;
    ThreadSettings.AutoFreeResults := FAutoFreeResults;
    ThreadSettings.Parent := Self;
    SetLength(FThreads, FThreadCountUse);
    for i := Low(FThreads) to High(FThreads) do begin
        FThreads[i] := TMultiThreadThread.Create(ThreadSettings);
    end;
end;

procedure TMultiThread2.Stop;
begin
    TerminateAllThreads;
end;

Constructor TMultiThread2.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    {$IFDEF MSWINDOWS}
    ThreadPriority := tpNormal;
    {$ENDIF}
    FThreadCountManual := 4;
    FAutoFreeData := True;
    FAutoFreeResults := True;
    FAutoFreeMessages := True;
    FAutoFreeParameters := True;
    WorkMessageQueue := TMultiThreadMessageQueue.Create;
    MessageQueue := TMultiThreadMessageQueue.Create;
    ResultMessageQueue := TMultiThreadMessageQueue.Create;
end;

Destructor TMultiThread2.Destroy;
begin
    Clear;
    Stop;
    if FAutoFreeParameters then begin
        if Assigned(Parameters) then begin
            FreeAndNil(Parameters);
        end;
    end;
    FreeAndNil(WorkMessageQueue);
    FreeAndNil(MessageQueue);
    FreeAndNil(ResultMessageQueue);
    inherited Destroy;
end;

function TMultiThread2.AddWork(Job: TObject): Boolean;
begin
    AtomicIncrement(FWorkToDo);
    WorkMessageQueue.Add(Job);
    Result := True;
end;

function TMultiThread2.Active: Boolean;
begin
    Result := Length(FThreads) > 0;
end;

procedure TMultiThread2.TerminateAllThreads;
var
    i: Integer;
begin
    for i := Low(FThreads) to High(FThreads) do begin
        FThreads[i].Terminate;
    end;
    for i := Low(FThreads) to High(FThreads) do begin
        FThreads[i].WaitFor;
        FreeAndNil(FThreads[i]);
    end;
    SetLength(FThreads, 0);
end;

procedure TMultiThread2.Clear;
var
    JobsFreed: Integer;
begin
    JobsFreed := WorkMessageQueue.Clear;
    MessageQueue.Clear;
    ResultMessageQueue.Clear;
    //FWorkToDo := 0;
    AtomicDecrement(FWorkToDo, JobsFreed);
end;

function TMultiThread2.GetWorkLeft: NativeInt;
begin
    Result := FWorkToDo;
end;

function TMultiThread2.GetWorkFinished: Boolean;
begin
    Result := FWorkToDo = 0;
end;

procedure TMultiThread2.CheckIncomingMessages;
var
    Obj: TObject;
begin
    while MessageQueue.Count > 0 do begin
        Obj := MessageQueue.Pop;
        if Assigned(Obj) then begin
            if Assigned(EThreadMessage) then begin
                EThreadMessage(Self, Obj);
            end;
            if FAutoFreeMessages then begin
                FreeAndNil(Obj);
            end;
        end;
    end;
end;

procedure TMultiThread2.CheckIncomingResults;
var
    Obj: TObject;
begin
    while ResultMessageQueue.Count > 0 do begin
        Obj := ResultMessageQueue.Pop;
        if Assigned(Obj) then begin
            if Assigned(EThreadResultsCallback) then begin
                EThreadResultsCallback(Self, Obj);
            end;
            if FAutoFreeResults then begin
                FreeAndNil(Obj);
            end;
        end;
    end;
end;

end.
